#include "config.h"
#include <stdio.h>
#include <string.h>
#include "param.h"
#include "routing.h"
#include "game.h"
#include "stage.h"

static int home_x, home_y;

// 隣接セルをキューに入れていくから
// 同一のセルはたかだか4つまでしか同時にキューに入らない
#define SEARCH_QUEUE_SIZE (MAP_WIDTH*MAP_HEIGHT*4)
#define ROUTE_STACK_SIZE (MAP_WIDTH*MAP_HEIGHT)


static route_value_t value_map[MAP_HEIGHT][MAP_WIDTH];
static route_dir_t dir_map[MAP_HEIGHT][MAP_WIDTH];

// valueが更新された(その先の探索が必要な)セルのキュー
// ・探索中に同じセルの値をより小さい値で更新するため、
// 　キューに入るセルの延べ数は全セル数より多い
// ・スタックで深さ優先探索するよりキューで幅優先探索したほうが、
//   遠回りルートでの無駄なvalue更新が減って効率が良い
static int search_queue_x[SEARCH_QUEUE_SIZE];
static int search_queue_y[SEARCH_QUEUE_SIZE];
static int search_queue_head;
static int search_queue_tail;

static int route_stack_x[ROUTE_STACK_SIZE];
static int route_stack_y[ROUTE_STACK_SIZE];
static int route_stack_top;

static int reachable_food_queue_x[FOOD_NUM_MAX];
static int reachable_food_queue_y[FOOD_NUM_MAX];
static int reachable_food_queue_head;
static int reachable_food_queue_tail;



void clear_search_queue(void);
void enqueue_search_queue(int x, int y);
int dequeue_search_queue(int *x, int *y);
void clear_reachable_food_queue(void);
void enqueue_reachable_food_queue(int x, int y);
int dequeue_reachable_food_queue(int *x, int *y);



void expand_route_value(void);
void expand_dir_map(int fx, int fy);    

void routing_init(void){
}

void setup_routing(void){
    home_x = get_stage_home_x();
    home_y = get_stage_home_y();
    make_value_map();
    // 実用上は不要だが、rコマンドで見えてしまうのでdir_mapは初期状態でクリアしておく
    memset(dir_map, di_invalid, sizeof(char)*MAP_WIDTH*MAP_HEIGHT);
}

void clear_search_queue(void){
    search_queue_head = search_queue_tail = 0;
}
void enqueue_search_queue(int x, int y){
    search_queue_x[search_queue_tail] = x;
    search_queue_y[search_queue_tail] = y;
    search_queue_tail = (search_queue_tail + 1) % SEARCH_QUEUE_SIZE;
    // キューがあふれることはないはず
    assert(search_queue_head != search_queue_tail);
}

// キューが空でないなら*x, *yにポップした座標をセットし非0を返す
int dequeue_search_queue(int *x, int *y){
    if (search_queue_head != search_queue_tail){
        *x = search_queue_x[search_queue_head];
        *y = search_queue_y[search_queue_head];
        search_queue_head = (search_queue_head + 1) % SEARCH_QUEUE_SIZE;
        return 1;
    }
    else {
        return 0;
    }
}

void clear_route_stack(void){
    route_stack_top = 0;
}
void push_route_stack(int x, int y){
    route_stack_x[route_stack_top] = x;
    route_stack_y[route_stack_top] = y;
    route_stack_top++;
    // キューがあふれることはないはず
    assert(route_stack_top < ROUTE_STACK_SIZE);
}
int pop_route_stack(int *x, int *y){
    if (route_stack_top == 0){
        return 0;
    }
    else {
        route_stack_top--;
        *x = route_stack_x[route_stack_top];
        *y = route_stack_y[route_stack_top];
        return 1;
    }
}
void clear_reachable_food_queue(void){
    reachable_food_queue_head = reachable_food_queue_tail = 0;
}
void enqueue_reachable_food_queue(int x, int y){
    // キューがあふれることはないはず
    assert(reachable_food_queue_tail < FOOD_NUM_MAX);

    reachable_food_queue_x[reachable_food_queue_tail] = x;
    reachable_food_queue_y[reachable_food_queue_tail] = y;
    // マップ内の餌のうち到達可能なものが入るだけなので、
    // リングバッファにしなくてもあふれない
    reachable_food_queue_tail++;
}

int dequeue_reachable_food_queue(int *x, int *y){
    if (reachable_food_queue_head < reachable_food_queue_tail){
        *x = reachable_food_queue_x[reachable_food_queue_head];
        *y = reachable_food_queue_y[reachable_food_queue_head];
        reachable_food_queue_head++;
        return 1;
    }
    else {
        return 0;
    }
}
int get_reachable_food_num(void){
    return reachable_food_queue_tail-reachable_food_queue_head;
}

void set_route_value(int x, int y, route_value_t value){
    value_map[y][x] = value;    
}
route_value_t get_route_value(int x, int y){
    return value_map[y][x];
}
void set_route_dir(int x, int y, dir_t dir){
    printf("set route dir(%d, %d) %c\n", x, y, Dir2Chr(dir));
    dir_map[y][x] = dir;
}
dir_t get_route_dir(int x, int y){
    return dir_map[y][x];
}


void make_value_map(void){
    printf("make value map from (%d, %d)\n", home_x, home_y);

    memset(value_map, ROUTE_VALUE_MAX, sizeof(char)*MAP_WIDTH*MAP_HEIGHT);
    clear_search_queue();
    clear_reachable_food_queue();
    
    set_route_value(home_x, home_y, 0);
    enqueue_search_queue(home_x, home_y);
    expand_route_value();
}

void expand_route_value(void){
    int x, y, nx, ny;
    dir_t dir;
    route_value_t route_value;
    while (dequeue_search_queue(&x, &y)){
        route_value = get_route_value(x, y);
        //        printf("(%d, %d)[%d] ", x, y, route_value);
        for (dir = di_begin; dir < di_end; dir++){
            nx = x + OffsetX(dir);
            ny = y + OffsetY(dir);
            if (get_map_element(nx, ny) == me_wall ||
                get_map_element(nx, ny) == me_block ||
                get_map_element(nx, ny) == me_player ||
                get_map_element(nx, ny) == me_player_food)
                {
                    continue;
                }
            if (get_route_value(nx, ny) > route_value+1){
                set_route_value(nx, ny, route_value+1);
                enqueue_search_queue(nx, ny);
            }
        }
        if (get_map_element(x, y) == me_food){
            enqueue_reachable_food_queue(x, y);
        }
    }
}



void make_dir_map(void){
    int fx, fy;
    if (get_reachable_food_num()>0){
        dequeue_reachable_food_queue(&fx, &fy);
        printf("make dir map to (%d, %d)\n", fx, fy);
        memset(dir_map, di_invalid, sizeof(char)*MAP_WIDTH*MAP_HEIGHT);
        expand_dir_map(fx, fy);
        // 幅優先探索だから最初にみつけた餌が最も近い＆最短経路
        // よってただちにルートを作成し餌食べ行動に入る
        set_game_state(gs_growing);
    }
    else {
        printf("no food reachable. skip make_dir_map().\n");
    }
}


void expand_dir_map(int x, int y){
    int hx, hy, nx, ny;
    dir_t dir;
    route_value_t rv, nrv;

    printf("make route dir from (%d, %d)\n", x, y);
    hx = get_stage_home_x();
    hy = get_stage_home_y();
    rv = get_route_value(x, y);

    while (x != hx || y != hy){
        for (dir = di_begin; dir < di_end; dir++){
            nx = x + OffsetX(dir);
            ny = y + OffsetY(dir);
            nrv = get_route_value(nx, ny);
            if (nrv < rv){
                set_route_dir(nx, ny, ReverseDir(dir));
                x = nx;
                y = ny;
                rv = nrv;
                break;
            }
        }
    }
}
